/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.example.warehouse.manager.service;

import java.text.NumberFormat;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.bluestemsoftware.open.eoa.example.warehouse.manager.service.dao.CustomerOrderDAO;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.util.Constants;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.util.DOMUtils;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.vo.CustomerOrder;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.vo.Item;
import org.bluestemsoftware.specification.eoa.application.spring.MyOperation;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.w3c.dom.Element;

public class AvailabilityProcessor implements Service {

    private Map<String, Integer> inventory;

    private ShippingProcessor shippingProcessor;
    private CustomerOrderDAO customerOrderDAO;

    public void checkAvailability(MyOperation myOperation, ActionContext requestContext) {

        // using a message property, which maps a predefined query
        // string to a schema type (a complex type in this case),
        // we extract the 'items' fragment from message
        
        Message request = requestContext.getMessage();
        Element itemsEl = (Element)request.evaluateNode(Constants.ORDER_ITEMS);
        
        // run through the entire order and make sure items are
        // in stock. if not return an out of stock fault. else,
        // we 'promise' them to customer

        Set<Item> items = null;
        
        synchronized (inventory) {

            Set<Item> itemsOutOfStock = new HashSet<Item>();
            for (Element item : DOMUtils.getChildElements(itemsEl)) {
                String id = item.getAttribute("id");
                int quantity = Integer.parseInt(item.getAttribute("quantity"));
                if ((inventory.get(id) - quantity) < 0) {
                    itemsOutOfStock.add(new Item(id, (inventory.get(id) - quantity) * -1));
                }
            }

            // if at least one item is out of stock. return out of stock
            // fault to client. note that deployment contains a runtime
            // definition for fault message which has predefined the 
            // payload element, i.e. 'Items'. we copy each out of stock
            // item and return fault

            ActionContext fault = null;
            if (itemsOutOfStock.size() > 0) {
                fault = myOperation.createFaultAction(Constants.OUT_OF_STOCK_FAULT);
                Message fMessage = fault.getMessage();
                Element temp = (Element)fMessage.evaluateNode(Constants.ORDER_ITEMS);
                for (Item item : itemsOutOfStock) {
                    Element itemEl = fMessage.createElementNS(Constants.MESSAGE_TNS, "Item"); 
                    itemEl.setAttribute("id", item.getId());
                    itemEl.setAttribute("quantity", Integer.toString(item.getQuantity()));
                    temp.appendChild(itemEl);
                }
                myOperation.sendResponse(fault);
                return;
            }

            // the order is available to promise. run through the request
            // again, decrement inventory on hand, and create a new order

            items = new HashSet<Item>();
            for (Element item : DOMUtils.getChildElements(itemsEl)) {
                String id = item.getAttribute("id");
                int quantity = Integer.parseInt(item.getAttribute("quantity"));
                inventory.put(id, inventory.get(id) - quantity);
                items.add(new Item(id, quantity));
            }

        }

        // get shipping cost estimate from shipping processor which
        // enapsulates shipping logic and use cost to create order

        String zip = request.evaluateString(Constants.ZIP_CODE);
        double sc = shippingProcessor.calculateShipping(zip, items);
        CustomerOrder c = customerOrderDAO.createCustomerOrder(zip, items, sc);

        // return the availability response to caller which implies
        // items are 'available to promise' and completes the message
        // exchange. note that runtime message definition is provided
        // within example-business-docs deployment descriptor (which
        // can be overridden within our application deployment if
        // desired). this prevents us from having to build msg up
        // manually, which is convenient for more complex structures

        ActionContext response = myOperation.createResponseAction();
        Element responsePayload = response.getMessage().getContent();
        Element shippingCostElement = DOMUtils.getFirstChildElement(responsePayload);
        String text = NumberFormat.getCurrencyInstance().format(sc);
        DOMUtils.setText(shippingCostElement, text);
        responsePayload.setAttribute("orderNumber", c.getOrderNumber());
        myOperation.sendResponse(response);

    }

    /*
     * dependency injector. set by spring framework
     */
    public void setInventory(Map<String, Integer> inventory) {
        this.inventory = inventory;
    }

    /*
     * dependency injector. set by spring framework
     */
    public void setShippingProcessor(ShippingProcessor shippingProcessor) {
        this.shippingProcessor = shippingProcessor;
    }

    /*
     * dependency injector. set by spring framework
     */
    public void setCustomerOrderDAO(CustomerOrderDAO customerOrderDAO) {
        this.customerOrderDAO = customerOrderDAO;
    }

    /*
     * Life cycle method invoked by spring framework
     */
    public void init() {
   
        // the application is 'wired' together by a third party
        // at deployment time. perform some sanity checks

        assert (shippingProcessor != null);
        assert (customerOrderDAO != null);
        assert (inventory != null);

    }

    /*
     * Life cycle method invoked by spring framework
     */
    public void destroy() {
        inventory.clear();
    }

}
